use std::num::NonZeroU32;

use crate::types::ObjectId;

/// Object Manager for creating, removing, and maintaining Wayland Objects
///
/// `T` should be a simple enum, like so:
/// ```
/// enum WaylandProtocol {
///     Display,
///     Registry,
///     LayerShell,
///     //...
/// }
/// ```
///
/// **It is important not to associate any data with this enum type, as it will greatly increase
/// our memory requirements for storing it.** To drive this point home we demand the users to
/// implement `Copy`, implying this type must be simply enough to be trivially copiable.
///
/// ### Why do it like this?
///
/// This is the lowest-overhead solution I could come up with. If the applicaiton developer knows
/// exactly with which Wayland objects it will interact, we have to need to create a generic object
/// manager that can deal with every possible use case. We can just make one that will deal with
/// just the objects that the developer actually cares about.
pub struct ObjectManager<T: Copy + PartialEq> {
    /// stores the object types. The position in this vector + 2 is the object id
    /// for example, if objects[0] == LayerSurface, then the object of id 2 is of
    /// the type "LayerSurface". This is because 0 is null and 1 is the display.
    objects: Vec<Option<T>>,
    /// the next id we ought to generate
    next: u32,
}

impl<T: Copy + PartialEq> ObjectManager<T> {
    /// Note, to instantiate an ObjectManager, you must pass in the type variant that represents
    /// the display. For example:
    /// ```
    /// #[derive(Clone, Copy, PartialEq)]
    /// enum WaylandProtocol {
    ///     Display,
    ///     Registry,
    ///     LayerShell,
    ///     //...
    /// }
    /// let objman = waybackend::objman::ObjectManager::new(WaylandProtocol::Display);
    /// ```
    ///
    /// NOTE THIS ISN'T TYPE CHECKED, AND **WILL** CAUSE PROBLEMS IF YOU DO IT WRONG.
    #[inline]
    pub fn new(display: T) -> Self {
        Self {
            objects: vec![Some(display)],
            next: 1,
        }
    }

    /// Returns the first objectId of the specified variant
    #[inline]
    #[must_use]
    pub fn get_first(&self, variant: T) -> Option<ObjectId> {
        self.objects
            .iter()
            .flatten()
            .position(|obj| *obj == variant)
            .map(|i| ObjectId::new(NonZeroU32::new(i as u32 + 1).unwrap()))
    }

    /// Gets all the objects that match the specified variant
    ///
    /// You can use this to, for example, find all binded outputs after initialization.
    #[inline]
    pub fn get_all(&self, variant: T) -> impl Iterator<Item = ObjectId> {
        self.objects
            .iter()
            .enumerate()
            .filter(move |(_, obj)| match obj {
                Some(obj) => *obj == variant,
                None => false,
            })
            .map(|(i, _)| ObjectId::new(NonZeroU32::new(i as u32 + 1).unwrap()))
    }

    /// get the type of the wayland object from its id
    ///
    /// Returns
    ///   * 'Some(T)' if the object still exists
    ///   * 'None' if the object was already deleted
    #[inline]
    #[must_use]
    pub fn get(&self, object_id: ObjectId) -> Option<T> {
        let pos = object_id.get().get() as usize - 1;
        self.objects[pos]
    }

    /// creates a new Id to use in requests
    #[inline]
    #[must_use]
    pub fn create(&mut self, object: T) -> ObjectId {
        if self.next as usize == self.objects.len() {
            self.next += 1;
            self.objects.push(Some(object));
            // SAFETY: we just added 1 above
            ObjectId::new(unsafe { NonZeroU32::new_unchecked(self.next) })
        } else {
            let i = self.next as usize;
            self.objects[i] = Some(object);

            // update next to the next available id
            self.next += 1;
            while (self.next as usize) < self.objects.len() {
                if self.objects[self.next as usize].is_none() {
                    break;
                }
                self.next += 1;
            }

            // SAFETY: we just added 1 above
            ObjectId::new(unsafe { NonZeroU32::new_unchecked(i as u32 + 1) })
        }
    }

    /// removes the wayland object.
    ///
    /// Removing the same element twice currently works just fine and does not panic,
    /// but that may change in the future
    #[inline]
    pub fn remove(&mut self, object_id: ObjectId) {
        let pos = object_id.get().get() - 1;
        self.objects[pos as usize] = None;
        if pos < self.next {
            self.next = pos;
        }
    }
}

#[cfg(test)]
mod tests {

    use super::*;

    #[derive(Clone, Copy, PartialEq)]
    enum DummyProtocol {
        Display,
        Region,
        Surface,
    }

    fn obj_from_u32(u: u32) -> ObjectId {
        ObjectId::new(NonZeroU32::new(u).unwrap())
    }

    #[test]
    fn creating_object_ids() {
        let mut manager = ObjectManager::new(DummyProtocol::Display);
        let id1 = manager.create(DummyProtocol::Region);
        assert_eq!(id1, obj_from_u32(2));
        let id2 = manager.create(DummyProtocol::Region);
        assert_eq!(id2, obj_from_u32(3));
        let id3 = manager.create(DummyProtocol::Region);
        assert_eq!(id3, obj_from_u32(4));

        manager.remove(id2);
        let id4 = manager.create(DummyProtocol::Region);
        assert_eq!(id4, id2);

        manager.remove(id1);
        let id5 = manager.create(DummyProtocol::Region);
        assert_eq!(id5, id1);

        manager.remove(id2);
        manager.remove(id1);
        let id6 = manager.create(DummyProtocol::Region);
        assert_eq!(id6, id1);

        let id7 = manager.create(DummyProtocol::Region);
        assert_eq!(id7, id2);
    }

    #[test]
    fn get_all() {
        let mut manager = ObjectManager::new(DummyProtocol::Display);
        let id1 = manager.create(DummyProtocol::Region);
        let id2 = manager.create(DummyProtocol::Surface);
        let id3 = manager.create(DummyProtocol::Region);
        let id4 = manager.create(DummyProtocol::Region);
        let id5 = manager.create(DummyProtocol::Surface);
        let id6 = manager.create(DummyProtocol::Surface);
        let id7 = manager.create(DummyProtocol::Region);
        let id8 = manager.create(DummyProtocol::Region);
        let id9 = manager.create(DummyProtocol::Region);
        let id10 = manager.create(DummyProtocol::Surface);

        let regions: Vec<ObjectId> = manager.get_all(DummyProtocol::Region).collect();
        let surfaces: Vec<ObjectId> = manager.get_all(DummyProtocol::Surface).collect();

        assert_eq!(&regions, &[id1, id3, id4, id7, id8, id9]);
        assert_eq!(&surfaces, &[id2, id5, id6, id10]);
    }
}
